@@ -1,7 +1,7 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Philippe Biondi <phil@secdev.org>
-# This program is published under a GPLv2 license
 
 """
 Implementation of the configuration object.
@@ -25,10 +25,11 @@ from scapy import VERSION
 from scapy.base_classes import BasePacket
 from scapy.consts import DARWIN, WINDOWS, LINUX, BSD, SOLARIS
 from scapy.error import log_scapy, warning, ScapyInvalidPlatformException
-from scapy.modules import six
+from scapy.libs import six
 from scapy.themes import NoTheme, apply_ipython_style
 
 from scapy.compat import (
+    cast,
     Any,
     Callable,
     DecoratorCallable,
@@ -38,16 +39,22 @@ from scapy.compat import (
     NoReturn,
     Optional,
     Set,
-    Type,
     Tuple,
+    Type,
     Union,
+    overload,
     TYPE_CHECKING,
 )
 from types import ModuleType
 
 if TYPE_CHECKING:
     # Do not import at runtime
+    import scapy.as_resolvers
+    from scapy.nmap import NmapKnowledgeBase
     from scapy.packet import Packet
+    from scapy.supersocket import SuperSocket  # noqa: F401
+    import scapy.asn1.asn1
+    import scapy.asn1.mib
 
 ############
 #  Config  #
@@ -208,7 +215,17 @@ class Num2Layer:
         # type: (int, Type[Packet]) -> None
         self.layer2num[layer] = num
 
+    @overload
     def __getitem__(self, item):
+        # type: (Type[Packet]) -> int
+        pass
+
+    @overload
+    def __getitem__(self, item):  # noqa: F811
+        # type: (int) -> Type[Packet]
+        pass
+
+    def __getitem__(self, item):  # noqa: F811
         # type: (Union[int, Type[Packet]]) -> Union[int, Type[Packet]]
         if isinstance(item, int):
             return self.num2layer[item]
@@ -272,6 +289,10 @@ class LayersList(List[Type['scapy.packet.Packet']]):
         result = []
         # This import may feel useless, but it is required for the eval below
         import scapy  # noqa: F401
+        try:
+            import builtins  # noqa: F401
+        except ImportError:
+            import __builtin__  # noqa: F401
         for lay in self.ldict:
             doc = eval(lay).__doc__
             result.append((lay, doc.strip().split("\n")[0] if doc else lay))
@@ -308,8 +329,9 @@ class CommandsList(List[Callable[..., Any]]):
         # type: () -> str
         s = []
         for li in sorted(self, key=lambda x: x.__name__):
-            doc = li.__doc__.split("\n")[0] if li.__doc__ else "--"
-            s.append("%-20s: %s" % (li.__name__, doc))
+            doc = li.__doc__ if li.__doc__ else "--"
+            doc = doc.lstrip().split('\n', 1)[0]
+            s.append("%-22s: %s" % (li.__name__, doc))
         return "\n".join(s)
 
     def register(self, cmd):
@@ -611,8 +633,13 @@ def _set_conf_sockets():
     if LINUX:
         from scapy.arch.linux import L3PacketSocket, L2Socket, L2ListenSocket
         conf.L3socket = L3PacketSocket
-        conf.L3socket6 = functools.partial(  # type: ignore
-            L3PacketSocket, filter="ip6")
+        conf.L3socket6 = cast(
+            "Type[SuperSocket]",
+            functools.partial(
+                L3PacketSocket,
+                filter="ip6"
+            )
+        )
         conf.L2socket = L2Socket
         conf.L2listen = L2ListenSocket
         conf.ifaces.reload()
@@ -627,10 +654,11 @@ def _set_conf_sockets():
         conf.ifaces.reload()
         # No need to update globals on Windows
         return
-    from scapy.supersocket import L3RawSocket
-    from scapy.layers.inet6 import L3RawSocket6
-    conf.L3socket = L3RawSocket
-    conf.L3socket6 = L3RawSocket6
+    else:
+        from scapy.supersocket import L3RawSocket
+        from scapy.layers.inet6 import L3RawSocket6
+        conf.L3socket = L3RawSocket
+        conf.L3socket6 = L3RawSocket6
 
 
 def _socket_changer(attr, val, old):
@@ -694,8 +722,10 @@ class Conf(ConfClass):
     iface = Interceptor("iface", None, _iface_changer)  # type: 'scapy.interfaces.NetworkInterface'  # type: ignore  # noqa: E501
     layers = LayersList()
     commands = CommandsList()  # type: CommandsList
-    ASN1_default_codec = None  #: Codec used by default for ASN1 objects
-    AS_resolver = None  #: choose the AS resolver class to use
+    #: Codec used by default for ASN1 objects
+    ASN1_default_codec = None  # type: 'scapy.asn1.asn1.ASN1Codec'
+    #: choose the AS resolver class to use
+    AS_resolver = None  # type: scapy.as_resolvers.AS_resolver
     dot15d4_protocol = None  # Used in dot15d4.py
     logLevel = Interceptor("logLevel", log_scapy.level, _loglevel_changer)
     #: if 0, doesn't check that IPID matches between IP sent and
@@ -719,9 +749,10 @@ class Conf(ConfClass):
     #: default mode for listening socket (to get answers if you
     #: spoof on a lan)
     promisc = True
-    sniff_promisc = 1  #: default mode for sniff()
+    #: default mode for sniff()
+    sniff_promisc = True  # type: bool
     raw_layer = None  # type: Type[Packet]
-    raw_summary = False
+    raw_summary = False  # type: Union[bool, Callable[[bytes], Any]]
     padding_layer = None  # type: Type[Packet]
     default_l2 = None  # type: Type[Packet]
     l2types = Num2Layer()
@@ -755,7 +786,7 @@ class Conf(ConfClass):
     #: holds the Scapy interface list and manager
     ifaces = None  # type: 'scapy.interfaces.NetworkInterfaceDict'
     #: holds the cache of interfaces loaded from Libpcap
-    cache_pcapiflist = {}  # type: Dict[str, Tuple[str, List[str], int]]
+    cache_pcapiflist = {}  # type: Dict[str, Tuple[str, List[str], Any, str]]
     neighbor = None  # type: 'scapy.layers.l2.Neighbor'
     # `neighbor` will be filed by scapy.layers.l2
     #: holds the Scapy IPv4 routing table and provides methods to
@@ -805,6 +836,7 @@ class Conf(ConfClass):
     load_layers = [
         'bluetooth',
         'bluetooth4LE',
+        'dcerpc',
         'dhcp',
         'dhcp6',
         'dns',
@@ -818,14 +850,18 @@ class Conf(ConfClass):
         'ipsec',
         'ir',
         'isakmp',
+        'kerberos',
         'l2',
         'l2tp',
+        'ldap',
         'llmnr',
         'lltd',
         'mgcp',
         'mobileip',
+        'mspac',
         'netbios',
         'netflow',
+        'ntlm',
         'ntp',
         'ppi',
         'ppp',
@@ -838,6 +874,8 @@ class Conf(ConfClass):
         'skinny',
         'smb',
         'smb2',
+        'smbclient',
+        'smbserver',
         'snmp',
         'tftp',
         'vrrp',
@@ -859,6 +897,8 @@ class Conf(ConfClass):
     #: Default is False.
     raise_no_dst_mac = False
     loopback_name = "lo" if LINUX else "lo0"
+    nmap_base = ""  # type: str
+    nmap_kdb = None  # type: Optional[NmapKnowledgeBase]
 
     def __getattribute__(self, attr):
         # type: (str) -> Any
@@ -878,6 +918,9 @@ class Conf(ConfClass):
         if attr == "services_tcp":
             from scapy.data import TCP_SERVICES
             return TCP_SERVICES
+        if attr == "services_sctp":
+            from scapy.data import SCTP_SERVICES
+            return SCTP_SERVICES
         if attr == "iface6":
             warnings.warn(
                 "conf.iface6 is deprecated in favor of conf.iface",
