@@ -1,7 +1,7 @@
+# SPDX-License-Identifier: GPL-2.0-only
 # This file is part of Scapy
-# See http://www.secdev.org/projects/scapy for more information
+# See https://scapy.net/ for more information
 # Copyright (C) Nils Weiss <nils@we155.de>
-# This program is published under a GPLv2 license
 
 # scapy.contrib.description = CANSocket Utils
 # scapy.contrib.status = loads
@@ -13,7 +13,7 @@ CANSocket.
 from scapy.error import log_loading
 from scapy.consts import LINUX
 from scapy.config import conf
-import scapy.modules.six as six
+import scapy.libs.six as six
 
 PYTHON_CAN = False
 
@@ -21,8 +21,6 @@ try:
     if conf.contribs['CANSocket']['use-python-can']:
         from can import BusABC as can_BusABC    # noqa: F401
         PYTHON_CAN = True
-    else:
-        PYTHON_CAN = False
 except ImportError:
     log_loading.info("Can't import python-can.")
 except KeyError:
@@ -30,18 +28,12 @@ except KeyError:
 
 
 if PYTHON_CAN:
-    log_loading.info("Using python-can CANSocket.")
-    log_loading.info("Specify 'conf.contribs['CANSocket'] = "
-                     "{'use-python-can': False}' to enable native CANSockets.")
+    log_loading.info("Using python-can CANSockets.\nSpecify 'conf.contribs['CANSocket'] = {'use-python-can': False}' to enable native CANSockets.")  # noqa: E501
     from scapy.contrib.cansocket_python_can import (PythonCANSocket, CANSocket)  # noqa: E501 F401
 
 elif LINUX and six.PY3 and not conf.use_pypy:
-    log_loading.info("Using native CANSocket.")
-    log_loading.info("Specify 'conf.contribs['CANSocket'] = "
-                     "{'use-python-can': True}' "
-                     "to enable python-can CANSockets.")
+    log_loading.info("Using native CANSockets.\nSpecify 'conf.contribs['CANSocket'] = {'use-python-can': True}' to enable python-can CANSockets.")  # noqa: E501
     from scapy.contrib.cansocket_native import (NativeCANSocket, CANSocket)  # noqa: E501 F401
 
 else:
-    log_loading.info("No CAN support available. Install python-can or "
-                     "use Linux and python3.")
+    log_loading.info("No CAN support available. Install python-can or use Linux and python3.")  # noqa: E501
